#                     GNU AFFERO GENERAL PUBLIC LICENSE
#                        Version 3, 19 November 2007
#
#  Copyright (C) 2007 Free Software Foundation, Inc. <https://fsf.org/>
#  Everyone is permitted to copy and distribute verbatim copies
#  of this license document, but changing it is not allowed.
#
#                             Preamble
#
#   The GNU Affero General Public License is a free, copyleft license for
# software and other kinds of works, specifically designed to ensure
# cooperation with the community in the case of network server software.
#
#   The licenses for most software and other practical works are designed
# to take away your freedom to share and change the works.  By contrast,
# our General Public Licenses are intended to guarantee your freedom to
# share and change all versions of a program--to make sure it remains free
# software for all its users.
#
#   When we speak of free software, we are referring to freedom, not
# price.  Our General Public Licenses are designed to make sure that you
# have the freedom to distribute copies of free software (and charge for
# them if you wish), that you receive source code or can get it if you
# want it, that you can change the software or use pieces of it in new
# free programs, and that you know you can do these things.
#
#   Developers that use our General Public Licenses protect your rights
# with two steps: (1) assert copyright on the software, and (2) offer
# you this License which gives you legal permission to copy, distribute
# and/or modify the software.
#
#   A secondary benefit of defending all users' freedom is that
# improvements made in alternate versions of the program, if they
# receive widespread use, become available for other developers to
# incorporate.  Many developers of free software are heartened and
# encouraged by the resulting cooperation.  However, in the case of
# software used on network servers, this result may fail to come about.
# The GNU General Public License permits making a modified version and
# letting the public access it on a server without ever releasing its
# source code to the public.
#
#   The GNU Affero General Public License is designed specifically to
# ensure that, in such cases, the modified source code becomes available
# to the community.  It requires the operator of a network server to
# provide the source code of the modified version running there to the
# users of that server.  Therefore, public use of a modified version, on
# a publicly accessible server, gives the public access to the source
# code of the modified version.
#
#   An older license, called the Affero General Public License and
# published by Affero, was designed to accomplish similar goals.  This is
# a different license, not a version of the Affero GPL, but Affero has
# released a new version of the Affero GPL which permits relicensing under
# this license.
#
#   The precise terms and conditions for copying, distribution and
# modification follow.
#
#                        TERMS AND CONDITIONS
#
#   0. Definitions.
#
#   "This License" refers to version 3 of the GNU Affero General Public License.
#
#   "Copyright" also means copyright-like laws that apply to other kinds of
# works, such as semiconductor masks.
#
#   "The Program" refers to any copyrightable work licensed under this
# License.  Each licensee is addressed as "you".  "Licensees" and
# "recipients" may be individuals or organizations.
#
#   To "modify" a work means to copy from or adapt all or part of the work
# in a fashion requiring copyright permission, other than the making of an
# exact copy.  The resulting work is called a "modified version" of the
# earlier work or a work "based on" the earlier work.
#
#   A "covered work" means either the unmodified Program or a work based
# on the Program.
#
#   To "propagate" a work means to do anything with it that, without
# permission, would make you directly or secondarily liable for
# infringement under applicable copyright law, except executing it on a
# computer or modifying a private copy.  Propagation includes copying,
# distribution (with or without modification), making available to the
# public, and in some countries other activities as well.
#
#   To "convey" a work means any kind of propagation that enables other
# parties to make or receive copies.  Mere interaction with a user through
# a computer network, with no transfer of a copy, is not conveying.
#
#   An interactive user interface displays "Appropriate Legal Notices"
# to the extent that it includes a convenient and prominently visible
# feature that (1) displays an appropriate copyright notice, and (2)
# tells the user that there is no warranty for the work (except to the
# extent that warranties are provided), that licensees may convey the
# work under this License, and how to view a copy of this License.  If
# the interface presents a list of user commands or options, such as a
# menu, a prominent item in the list meets this criterion.
#
#   1. Source Code.
#
#   The "source code" for a work means the preferred form of the work
# for making modifications to it.  "Object code" means any non-source
# form of a work.
#
#   A "Standard Interface" means an interface that either is an official
# standard defined by a recognized standards body, or, in the case of
# interfaces specified for a particular programming language, one that
# is widely used among developers working in that language.
#
#   The "System Libraries" of an executable work include anything, other
# than the work as a whole, that (a) is included in the normal form of
# packaging a Major Component, but which is not part of that Major
# Component, and (b) serves only to enable use of the work with that
# Major Component, or to implement a Standard Interface for which an
# implementation is available to the public in source code form.  A
# "Major Component", in this context, means a major essential component
# (kernel, window system, and so on) of the specific operating system
# (if any) on which the executable work runs, or a compiler used to
# produce the work, or an object code interpreter used to run it.
#
#   The "Corresponding Source" for a work in object code form means all
# the source code needed to generate, install, and (for an executable
# work) run the object code and to modify the work, including scripts to
# control those activities.  However, it does not include the work's
# System Libraries, or general-purpose tools or generally available free
# programs which are used unmodified in performing those activities but
# which are not part of the work.  For example, Corresponding Source
# includes interface definition files associated with source files for
# the work, and the source code for shared libraries and dynamically
# linked subprograms that the work is specifically designed to require,
# such as by intimate data communication or control flow between those
# subprograms and other parts of the work.
#
#   The Corresponding Source need not include anything that users
# can regenerate automatically from other parts of the Corresponding
# Source.
#
#   The Corresponding Source for a work in source code form is that
# same work.
#
#   2. Basic Permissions.
#
#   All rights granted under this License are granted for the term of
# copyright on the Program, and are irrevocable provided the stated
# conditions are met.  This License explicitly affirms your unlimited
# permission to run the unmodified Program.  The output from running a
# covered work is covered by this License only if the output, given its
# content, constitutes a covered work.  This License acknowledges your
# rights of fair use or other equivalent, as provided by copyright law.
#
#   You may make, run and propagate covered works that you do not
# convey, without conditions so long as your license otherwise remains
# in force.  You may convey covered works to others for the sole purpose
# of having them make modifications exclusively for you, or provide you
# with facilities for running those works, provided that you comply with
# the terms of this License in conveying all material for which you do
# not control copyright.  Those thus making or running the covered works
# for you must do so exclusively on your behalf, under your direction
# and control, on terms that prohibit them from making any copies of
# your copyrighted material outside their relationship with you.
#
#   Conveying under any other circumstances is permitted solely under
# the conditions stated below.  Sublicensing is not allowed; section 10
# makes it unnecessary.
#
#   3. Protecting Users' Legal Rights From Anti-Circumvention Law.
#
#   No covered work shall be deemed part of an effective technological
# measure under any applicable law fulfilling obligations under article
# 11 of the WIPO copyright treaty adopted on 20 December 1996, or
# similar laws prohibiting or restricting circumvention of such
# measures.
#
#   When you convey a covered work, you waive any legal power to forbid
# circumvention of technological measures to the extent such circumvention
# is effected by exercising rights under this License with respect to
# the covered work, and you disclaim any intention to limit operation or
# modification of the work as a means of enforcing, against the work's
# users, your or third parties' legal rights to forbid circumvention of
# technological measures.
#
#   4. Conveying Verbatim Copies.
#
#   You may convey verbatim copies of the Program's source code as you
# receive it, in any medium, provided that you conspicuously and
# appropriately publish on each copy an appropriate copyright notice;
# keep intact all notices stating that this License and any
# non-permissive terms added in accord with section 7 apply to the code;
# keep intact all notices of the absence of any warranty; and give all
# recipients a copy of this License along with the Program.
#
#   You may charge any price or no price for each copy that you convey,
# and you may offer support or warranty protection for a fee.
#
#   5. Conveying Modified Source Versions.
#
#   You may convey a work based on the Program, or the modifications to
# produce it from the Program, in the form of source code under the
# terms of section 4, provided that you also meet all of these conditions:
#
#     a) The work must carry prominent notices stating that you modified
#     it, and giving a relevant date.
#
#     b) The work must carry prominent notices stating that it is
#     released under this License and any conditions added under section
#     7.  This requirement modifies the requirement in section 4 to
#     "keep intact all notices".
#
#     c) You must license the entire work, as a whole, under this
#     License to anyone who comes into possession of a copy.  This
#     License will therefore apply, along with any applicable section 7
#     additional terms, to the whole of the work, and all its parts,
#     regardless of how they are packaged.  This License gives no
#     permission to license the work in any other way, but it does not
#     invalidate such permission if you have separately received it.
#
#     d) If the work has interactive user interfaces, each must display
#     Appropriate Legal Notices; however, if the Program has interactive
#     interfaces that do not display Appropriate Legal Notices, your
#     work need not make them do so.
#
#   A compilation of a covered work with other separate and independent
# works, which are not by their nature extensions of the covered work,
# and which are not combined with it such as to form a larger program,
# in or on a volume of a storage or distribution medium, is called an
# "aggregate" if the compilation and its resulting copyright are not
# used to limit the access or legal rights of the compilation's users
# beyond what the individual works permit.  Inclusion of a covered work
# in an aggregate does not cause this License to apply to the other
# parts of the aggregate.
#
#   6. Conveying Non-Source Forms.
#
#   You may convey a covered work in object code form under the terms
# of sections 4 and 5, provided that you also convey the
# machine-readable Corresponding Source under the terms of this License,
# in one of these ways:
#
#     a) Convey the object code in, or embodied in, a physical product
#     (including a physical distribution medium), accompanied by the
#     Corresponding Source fixed on a durable physical medium
#     customarily used for software interchange.
#
#     b) Convey the object code in, or embodied in, a physical product
#     (including a physical distribution medium), accompanied by a
#     written offer, valid for at least three years and valid for as
#     long as you offer spare parts or customer support for that product
#     model, to give anyone who possesses the object code either (1) a
#     copy of the Corresponding Source for all the software in the
#     product that is covered by this License, on a durable physical
#     medium customarily used for software interchange, for a price no
#     more than your reasonable cost of physically performing this
#     conveying of source, or (2) access to copy the
#     Corresponding Source from a network server at no charge.
#
#     c) Convey individual copies of the object code with a copy of the
#     written offer to provide the Corresponding Source.  This
#     alternative is allowed only occasionally and noncommercially, and
#     only if you received the object code with such an offer, in accord
#     with subsection 6b.
#
#     d) Convey the object code by offering access from a designated
#     place (gratis or for a charge), and offer equivalent access to the
#     Corresponding Source in the same way through the same place at no
#     further charge.  You need not require recipients to copy the
#     Corresponding Source along with the object code.  If the place to
#     copy the object code is a network server, the Corresponding Source
#     may be on a different server (operated by you or a third party)
#     that supports equivalent copying facilities, provided you maintain
#     clear directions next to the object code saying where to find the
#     Corresponding Source.  Regardless of what server hosts the
#     Corresponding Source, you remain obligated to ensure that it is
#     available for as long as needed to satisfy these requirements.
#
#     e) Convey the object code using peer-to-peer transmission, provided
#     you inform other peers where the object code and Corresponding
#     Source of the work are being offered to the general public at no
#     charge under subsection 6d.
#
#   A separable portion of the object code, whose source code is excluded
# from the Corresponding Source as a System Library, need not be
# included in conveying the object code work.
#
#   A "User Product" is either (1) a "consumer product", which means any
# tangible personal property which is normally used for personal, family,
# or household purposes, or (2) anything designed or sold for incorporation
# into a dwelling.  In determining whether a product is a consumer product,
# doubtful cases shall be resolved in favor of coverage.  For a particular
# product received by a particular user, "normally used" refers to a
# typical or common use of that class of product, regardless of the status
# of the particular user or of the way in which the particular user
# actually uses, or expects or is expected to use, the product.  A product
# is a consumer product regardless of whether the product has substantial
# commercial, industrial or non-consumer uses, unless such uses represent
# the only significant mode of use of the product.
#
#   "Installation Information" for a User Product means any methods,
# procedures, authorization keys, or other information required to install
# and execute modified versions of a covered work in that User Product from
# a modified version of its Corresponding Source.  The information must
# suffice to ensure that the continued functioning of the modified object
# code is in no case prevented or interfered with solely because
# modification has been made.
#
#   If you convey an object code work under this section in, or with, or
# specifically for use in, a User Product, and the conveying occurs as
# part of a transaction in which the right of possession and use of the
# User Product is transferred to the recipient in perpetuity or for a
# fixed term (regardless of how the transaction is characterized), the
# Corresponding Source conveyed under this section must be accompanied
# by the Installation Information.  But this requirement does not apply
# if neither you nor any third party retains the ability to install
# modified object code on the User Product (for example, the work has
# been installed in ROM).
#
#   The requirement to provide Installation Information does not include a
# requirement to continue to provide support service, warranty, or updates
# for a work that has been modified or installed by the recipient, or for
# the User Product in which it has been modified or installed.  Access to a
# network may be denied when the modification itself materially and
# adversely affects the operation of the network or violates the rules and
# protocols for communication across the network.
#
#   Corresponding Source conveyed, and Installation Information provided,
# in accord with this section must be in a format that is publicly
# documented (and with an implementation available to the public in
# source code form), and must require no special password or key for
# unpacking, reading or copying.
#
#   7. Additional Terms.
#
#   "Additional permissions" are terms that supplement the terms of this
# License by making exceptions from one or more of its conditions.
# Additional permissions that are applicable to the entire Program shall
# be treated as though they were included in this License, to the extent
# that they are valid under applicable law.  If additional permissions
# apply only to part of the Program, that part may be used separately
# under those permissions, but the entire Program remains governed by
# this License without regard to the additional permissions.
#
#   When you convey a copy of a covered work, you may at your option
# remove any additional permissions from that copy, or from any part of
# it.  (Additional permissions may be written to require their own
# removal in certain cases when you modify the work.)  You may place
# additional permissions on material, added by you to a covered work,
# for which you have or can give appropriate copyright permission.
#
#   Notwithstanding any other provision of this License, for material you
# add to a covered work, you may (if authorized by the copyright holders of
# that material) supplement the terms of this License with terms:
#
#     a) Disclaiming warranty or limiting liability differently from the
#     terms of sections 15 and 16 of this License; or
#
#     b) Requiring preservation of specified reasonable legal notices or
#     author attributions in that material or in the Appropriate Legal
#     Notices displayed by works containing it; or
#
#     c) Prohibiting misrepresentation of the origin of that material, or
#     requiring that modified versions of such material be marked in
#     reasonable ways as different from the original version; or
#
#     d) Limiting the use for publicity purposes of names of licensors or
#     authors of the material; or
#
#     e) Declining to grant rights under trademark law for use of some
#     trade names, trademarks, or service marks; or
#
#     f) Requiring indemnification of licensors and authors of that
#     material by anyone who conveys the material (or modified versions of
#     it) with contractual assumptions of liability to the recipient, for
#     any liability that these contractual assumptions directly impose on
#     those licensors and authors.
#
#   All other non-permissive additional terms are considered "further
# restrictions" within the meaning of section 10.  If the Program as you
# received it, or any part of it, contains a notice stating that it is
# governed by this License along with a term that is a further
# restriction, you may remove that term.  If a license document contains
# a further restriction but permits relicensing or conveying under this
# License, you may add to a covered work material governed by the terms
# of that license document, provided that the further restriction does
# not survive such relicensing or conveying.
#
#   If you add terms to a covered work in accord with this section, you
# must place, in the relevant source files, a statement of the
# additional terms that apply to those files, or a notice indicating
# where to find the applicable terms.
#
#   Additional terms, permissive or non-permissive, may be stated in the
# form of a separately written license, or stated as exceptions;
# the above requirements apply either way.
#
#   8. Termination.
#
#   You may not propagate or modify a covered work except as expressly
# provided under this License.  Any attempt otherwise to propagate or
# modify it is void, and will automatically terminate your rights under
# this License (including any patent licenses granted under the third
# paragraph of section 11).
#
#   However, if you cease all violation of this License, then your
# license from a particular copyright holder is reinstated (a)
# provisionally, unless and until the copyright holder explicitly and
# finally terminates your license, and (b) permanently, if the copyright
# holder fails to notify you of the violation by some reasonable means
# prior to 60 days after the cessation.
#
#   Moreover, your license from a particular copyright holder is
# reinstated permanently if the copyright holder notifies you of the
# violation by some reasonable means, this is the first time you have
# received notice of violation of this License (for any work) from that
# copyright holder, and you cure the violation prior to 30 days after
# your receipt of the notice.
#
#   Termination of your rights under this section does not terminate the
# licenses of parties who have received copies or rights from you under
# this License.  If your rights have been terminated and not permanently
# reinstated, you do not qualify to receive new licenses for the same
# material under section 10.
#
#   9. Acceptance Not Required for Having Copies.
#
#   You are not required to accept this License in order to receive or
# run a copy of the Program.  Ancillary propagation of a covered work
# occurring solely as a consequence of using peer-to-peer transmission
# to receive a copy likewise does not require acceptance.  However,
# nothing other than this License grants you permission to propagate or
# modify any covered work.  These actions infringe copyright if you do
# not accept this License.  Therefore, by modifying or propagating a
# covered work, you indicate your acceptance of this License to do so.
#
#   10. Automatic Licensing of Downstream Recipients.
#
#   Each time you convey a covered work, the recipient automatically
# receives a license from the original licensors, to run, modify and
# propagate that work, subject to this License.  You are not responsible
# for enforcing compliance by third parties with this License.
#
#   An "entity transaction" is a transaction transferring control of an
# organization, or substantially all assets of one, or subdividing an
# organization, or merging organizations.  If propagation of a covered
# work results from an entity transaction, each party to that
# transaction who receives a copy of the work also receives whatever
# licenses to the work the party's predecessor in interest had or could
# give under the previous paragraph, plus a right to possession of the
# Corresponding Source of the work from the predecessor in interest, if
# the predecessor has it or can get it with reasonable efforts.
#
#   You may not impose any further restrictions on the exercise of the
# rights granted or affirmed under this License.  For example, you may
# not impose a license fee, royalty, or other charge for exercise of
# rights granted under this License, and you may not initiate litigation
# (including a cross-claim or counterclaim in a lawsuit) alleging that
# any patent claim is infringed by making, using, selling, offering for
# sale, or importing the Program or any portion of it.
#
#   11. Patents.
#
#   A "contributor" is a copyright holder who authorizes use under this
# License of the Program or a work on which the Program is based.  The
# work thus licensed is called the contributor's "contributor version".
#
#   A contributor's "essential patent claims" are all patent claims
# owned or controlled by the contributor, whether already acquired or
# hereafter acquired, that would be infringed by some manner, permitted
# by this License, of making, using, or selling its contributor version,
# but do not include claims that would be infringed only as a
# consequence of further modification of the contributor version.  For
# purposes of this definition, "control" includes the right to grant
# patent sublicenses in a manner consistent with the requirements of
# this License.
#
#   Each contributor grants you a non-exclusive, worldwide, royalty-free
# patent license under the contributor's essential patent claims, to
# make, use, sell, offer for sale, import and otherwise run, modify and
# propagate the contents of its contributor version.
#
#   In the following three paragraphs, a "patent license" is any express
# agreement or commitment, however denominated, not to enforce a patent
# (such as an express permission to practice a patent or covenant not to
# sue for patent infringement).  To "grant" such a patent license to a
# party means to make such an agreement or commitment not to enforce a
# patent against the party.
#
#   If you convey a covered work, knowingly relying on a patent license,
# and the Corresponding Source of the work is not available for anyone
# to copy, free of charge and under the terms of this License, through a
# publicly available network server or other readily accessible means,
# then you must either (1) cause the Corresponding Source to be so
# available, or (2) arrange to deprive yourself of the benefit of the
# patent license for this particular work, or (3) arrange, in a manner
# consistent with the requirements of this License, to extend the patent
# license to downstream recipients.  "Knowingly relying" means you have
# actual knowledge that, but for the patent license, your conveying the
# covered work in a country, or your recipient's use of the covered work
# in a country, would infringe one or more identifiable patents in that
# country that you have reason to believe are valid.
#
#   If, pursuant to or in connection with a single transaction or
# arrangement, you convey, or propagate by procuring conveyance of, a
# covered work, and grant a patent license to some of the parties
# receiving the covered work authorizing them to use, propagate, modify
# or convey a specific copy of the covered work, then the patent license
# you grant is automatically extended to all recipients of the covered
# work and works based on it.
#
#   A patent license is "discriminatory" if it does not include within
# the scope of its coverage, prohibits the exercise of, or is
# conditioned on the non-exercise of one or more of the rights that are
# specifically granted under this License.  You may not convey a covered
# work if you are a party to an arrangement with a third party that is
# in the business of distributing software, under which you make payment
# to the third party based on the extent of your activity of conveying
# the work, and under which the third party grants, to any of the
# parties who would receive the covered work from you, a discriminatory
# patent license (a) in connection with copies of the covered work
# conveyed by you (or copies made from those copies), or (b) primarily
# for and in connection with specific products or compilations that
# contain the covered work, unless you entered into that arrangement,
# or that patent license was granted, prior to 28 March 2007.
#
#   Nothing in this License shall be construed as excluding or limiting
# any implied license or other defenses to infringement that may
# otherwise be available to you under applicable patent law.
#
#   12. No Surrender of Others' Freedom.
#
#   If conditions are imposed on you (whether by court order, agreement or
# otherwise) that contradict the conditions of this License, they do not
# excuse you from the conditions of this License.  If you cannot convey a
# covered work so as to satisfy simultaneously your obligations under this
# License and any other pertinent obligations, then as a consequence you may
# not convey it at all.  For example, if you agree to terms that obligate you
# to collect a royalty for further conveying from those to whom you convey
# the Program, the only way you could satisfy both those terms and this
# License would be to refrain entirely from conveying the Program.
#
#   13. Remote Network Interaction; Use with the GNU General Public License.
#
#   Notwithstanding any other provision of this License, if you modify the
# Program, your modified version must prominently offer all users
# interacting with it remotely through a computer network (if your version
# supports such interaction) an opportunity to receive the Corresponding
# Source of your version by providing access to the Corresponding Source
# from a network server at no charge, through some standard or customary
# means of facilitating copying of software.  This Corresponding Source
# shall include the Corresponding Source for any work covered by version 3
# of the GNU General Public License that is incorporated pursuant to the
# following paragraph.
#
#   Notwithstanding any other provision of this License, you have
# permission to link or combine any covered work with a work licensed
# under version 3 of the GNU General Public License into a single
# combined work, and to convey the resulting work.  The terms of this
# License will continue to apply to the part which is the covered work,
# but the work with which it is combined will remain governed by version
# 3 of the GNU General Public License.
#
#   14. Revised Versions of this License.
#
#   The Free Software Foundation may publish revised and/or new versions of
# the GNU Affero General Public License from time to time.  Such new versions
# will be similar in spirit to the present version, but may differ in detail to
# address new problems or concerns.
#
#   Each version is given a distinguishing version number.  If the
# Program specifies that a certain numbered version of the GNU Affero General
# Public License "or any later version" applies to it, you have the
# option of following the terms and conditions either of that numbered
# version or of any later version published by the Free Software
# Foundation.  If the Program does not specify a version number of the
# GNU Affero General Public License, you may choose any version ever published
# by the Free Software Foundation.
#
#   If the Program specifies that a proxy can decide which future
# versions of the GNU Affero General Public License can be used, that proxy's
# public statement of acceptance of a version permanently authorizes you
# to choose that version for the Program.
#
#   Later license versions may give you additional or different
# permissions.  However, no additional obligations are imposed on any
# author or copyright holder as a result of your choosing to follow a
# later version.
#
#   15. Disclaimer of Warranty.
#
#   THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
# APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
# HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
# OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
# IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
# ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
#
#   16. Limitation of Liability.
#
#   IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
# WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
# THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
# GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
# USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
# DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
# PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
# EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGES.
#
#   17. Interpretation of Sections 15 and 16.
#
#   If the disclaimer of warranty and limitation of liability provided
# above cannot be given local legal effect according to their terms,
# reviewing courts shall apply local law that most closely approximates
# an absolute waiver of all civil liability in connection with the
# Program, unless a warranty or assumption of liability accompanies a
# copy of the Program in return for a fee.
#
#                      END OF TERMS AND CONDITIONS
#
#             How to Apply These Terms to Your New Programs
#
#   If you develop a new program, and you want it to be of the greatest
# possible use to the public, the best way to achieve this is to make it
# free software which everyone can redistribute and change under these terms.
#
#   To do so, attach the following notices to the program.  It is safest
# to attach them to the start of each source file to most effectively
# state the exclusion of warranty; and each file should have at least
# the "copyright" line and a pointer to where the full notice is found.
#
#     <one line to give the program's name and a brief idea of what it does.>
#     Copyright (C) <year>  <name of author>
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Also add information on how to contact you by electronic and paper mail.
#
#   If your software can interact with users remotely through a computer
# network, you should also make sure that it provides a way for users to
# get its source.  For example, if your program is a web application, its
# interface could display a "Source" link that leads users to an archive
# of the code.  There are many ways you could offer source, and different
# solutions will be better for different programs; see section 13 for the
# specific requirements.
#
#   You should also get your employer (if you work as a programmer) or school,
# if any, to sign a "copyright disclaimer" for the program, if necessary.
# For more information on this, and how to apply and follow the GNU AGPL, see
# <https://www.gnu.org/licenses/>.

# Open Stochastic Daily Unit Commitment of Thermal and ESS Units (openSDUC) - Version 1.3.32 - January 27, 2025
# simplicity and transparency in power systems planning

# Developed by

#    Andres Ramos
#    Instituto de Investigacion Tecnologica
#    Escuela Tecnica Superior de Ingenieria - ICAI
#    UNIVERSIDAD PONTIFICIA COMILLAS
#    Alberto Aguilera 23
#    28015 Madrid, Spain
#    Andres.Ramos@comillas.edu
#    https://pascua.iit.comillas.edu/aramos/Ramos_CV.htm

#%% Libraries
import argparse
import os
import pandas        as pd
import time          # count clock time
import psutil        # access the number of CPUs
import pyomo.environ as pyo
from   pyomo.environ import Set, Var, Binary, NonNegativeReals, RealSet, Constraint, ConcreteModel, Objective, minimize, Suffix, DataPortal
from   pyomo.opt     import SolverFactory

import matplotlib.pyplot as plt

print('\033[34m#### Academic research license - for non-commercial use only ####\033[0m')

StartTime = time.time()

parser = argparse.ArgumentParser(description='Introducing main parameters...')
parser.add_argument('--case',   type=str, default=None)
parser.add_argument('--dir',    type=str, default=None)
parser.add_argument('--solver', type=str, default=None)

DIR    = os.path.dirname(__file__)
CASE   = '16g'
SOLVER = 'gurobi'

def main():
    args = parser.parse_args()
    if args.dir is None:
        args.dir    = input('Input Dir    Name (Default {}): '.format(DIR))
        if args.dir == '':
            args.dir = DIR
    if args.case is None:
        args.case   = input('Input Case   Name (Default {}): '.format(CASE))
        if args.case == '':
            args.case = CASE
    if args.solver is None:
        args.solver = input('Input Solver Name (Default {}): '.format(SOLVER))
        if args.solver == '':
            args.solver = SOLVER
    print(args.case)
    print(args.dir)
    print(args.solver)
    import sys
    print(sys.argv)
    print(args)
    openSDUC_run(args.dir, args.case, args.solver)

def openSDUC_run(DirName, CaseName, SolverName):

    _path = os.path.join(DirName, CaseName)
    StartTime = time.time()

    #%% model declaration
    mSDUC = ConcreteModel('Open Stochastic Daily Unit Commitment of Thermal and ESS Units (openSDUC) - Version 1.3.32 - January 27, 2025')

    #%% reading the sets
    dictSets = DataPortal()
    dictSets.load(filename=_path+'/oUC_Dict_Scenario_'  +CaseName+'.csv', set='sc', format='set')
    dictSets.load(filename=_path+'/oUC_Dict_LoadLevel_' +CaseName+'.csv', set='n' , format='set')
    dictSets.load(filename=_path+'/oUC_Dict_Generation_'+CaseName+'.csv', set='g' , format='set')
    dictSets.load(filename=_path+'/oUC_Dict_Storage_'   +CaseName+'.csv', set='st', format='set')
    dictSets.load(filename=_path+'/oUC_Dict_Technology_'+CaseName+'.csv', set='gt', format='set')
    dictSets.load(filename=_path+'/oUC_Dict_Company_'   +CaseName+'.csv', set='co', format='set')

    mSDUC.sc = Set(initialize=dictSets['sc'], doc='scenarios'   )
    mSDUC.nn = Set(initialize=dictSets['n' ], doc='load levels' )
    mSDUC.gg = Set(initialize=dictSets['g' ], doc='units'       )
    mSDUC.gt = Set(initialize=dictSets['gt'], doc='technologies')
    mSDUC.co = Set(initialize=dictSets['co'], doc='companies'   )
    mSDUC.st = Set(initialize=dictSets['st'], doc='ESS types'   )

    #%% reading data from CSV files
    dfParameter          = pd.read_csv(_path+'/oUC_Data_Parameter_'        +CaseName+'.csv', index_col=[0  ])
    dfDuration           = pd.read_csv(_path+'/oUC_Data_Duration_'         +CaseName+'.csv', index_col=[0  ])
    dfScenario           = pd.read_csv(_path+'/oUC_Data_Scenario_'         +CaseName+'.csv', index_col=[0  ])
    dfDemand             = pd.read_csv(_path+'/oUC_Data_Demand_'           +CaseName+'.csv', index_col=[0,1])
    dfOperatingReserve   = pd.read_csv(_path+'/oUC_Data_OperatingReserve_' +CaseName+'.csv', index_col=[0,1])
    dfGeneration         = pd.read_csv(_path+'/oUC_Data_Generation_'       +CaseName+'.csv', index_col=[0  ])
    dfVariableMinPower   = pd.read_csv(_path+'/oUC_Data_MinimumGeneration_'+CaseName+'.csv', index_col=[0,1])
    dfVariableMaxPower   = pd.read_csv(_path+'/oUC_Data_MaximumGeneration_'+CaseName+'.csv', index_col=[0,1])
    dfVariableMinStorage = pd.read_csv(_path+'/oUC_Data_MinimumStorage_'   +CaseName+'.csv', index_col=[0,1])
    dfVariableMaxStorage = pd.read_csv(_path+'/oUC_Data_MaximumStorage_'   +CaseName+'.csv', index_col=[0,1])
    dfEnergyInflows      = pd.read_csv(_path+'/oUC_Data_EnergyInflows_'    +CaseName+'.csv', index_col=[0,1])

    # substitute NaN by 0
    dfParameter.fillna         (0.0, inplace=True)
    dfDuration.fillna          (0.0, inplace=True)
    dfScenario.fillna          (0.0, inplace=True)
    dfDemand.fillna            (0.0, inplace=True)
    dfOperatingReserve.fillna  (0.0, inplace=True)
    dfGeneration.fillna        (0.0, inplace=True)
    dfVariableMinPower.fillna  (0.0, inplace=True)
    dfVariableMaxPower.fillna  (0.0, inplace=True)
    dfVariableMinStorage.fillna(0.0, inplace=True)
    dfVariableMaxStorage.fillna(0.0, inplace=True)
    dfEnergyInflows.fillna     (0.0, inplace=True)

    #%% general parameters
    pENSCost            = dfParameter['ENSCost' ].iloc[0] * 1e-3                                                                   # cost of energy not served        [MEUR/GWh]
    pCO2Cost            = dfParameter['CO2Cost' ].iloc[0]                                                                          # cost of CO2 emission             [EUR/CO2 ton]
    pTimeStep           = dfParameter['TimeStep'].iloc[0].astype('int')                                                            # duration of the unit time step   [h]

    pDuration           = dfDuration          ['Duration'    ] * pTimeStep                                                         # duration of load levels          [h]
    pScenProb           = dfScenario          ['Probability' ]                                                                     # probabilities of scenarios       [p.u.]
    pDemand             = dfDemand            ['Demand'      ] * 1e-3                                                              # demand                           [GW]
    pOperReserveUp      = dfOperatingReserve  ['Up'          ] * 1e-3                                                              # operating reserve up             [GW]
    pOperReserveDw      = dfOperatingReserve  ['Down'        ] * 1e-3                                                              # operating reserve down           [GW]
    pVariableMinPower   = dfVariableMinPower  [mSDUC.gg]       * 1e-3                                                              # dynamic variable minimum power   [GW]
    pVariableMaxPower   = dfVariableMaxPower  [mSDUC.gg]       * 1e-3                                                              # dynamic variable maximum power   [GW]
    pVariableMinStorage = dfVariableMinStorage[mSDUC.gg]                                                                           # dynamic variable minimum storage [GWh]
    pVariableMaxStorage = dfVariableMaxStorage[mSDUC.gg]                                                                           # dynamic variable maximum storage [GWh]
    pEnergyInflows      = dfEnergyInflows     [mSDUC.gg]       * 1e-3                                                              # dynamic energy inflows           [GW]

    # compute the demand as the mean over the time step load levels and assign it to active load levels. Idem for operating reserve, variable max power, variable min and max storage capacity and inflows
    pDemand             = pDemand.rolling            (pTimeStep).mean()
    pOperReserveUp      = pOperReserveUp.rolling     (pTimeStep).mean()
    pOperReserveDw      = pOperReserveDw.rolling     (pTimeStep).mean()
    pVariableMinPower   = pVariableMinPower.rolling  (pTimeStep).mean()
    pVariableMaxPower   = pVariableMaxPower.rolling  (pTimeStep).mean()
    pVariableMinStorage = pVariableMinStorage.rolling(pTimeStep).mean()
    pVariableMaxStorage = pVariableMaxStorage.rolling(pTimeStep).mean()
    pEnergyInflows      = pEnergyInflows.rolling     (pTimeStep).mean()

    pDemand.fillna            (0.0, inplace=True)
    pOperReserveUp.fillna     (0.0, inplace=True)
    pOperReserveDw.fillna     (0.0, inplace=True)
    pVariableMinPower.fillna  (0.0, inplace=True)
    pVariableMaxPower.fillna  (0.0, inplace=True)
    pVariableMinStorage.fillna(0.0, inplace=True)
    pVariableMaxStorage.fillna(0.0, inplace=True)
    pEnergyInflows.fillna     (0.0, inplace=True)

    if pTimeStep > 1:
        # assign duration 0 to load levels not being considered, active load levels are at the end of every pTimeStep
        for i in range(pTimeStep-2,-1,-1):
            pDuration[range(i,len(mSDUC.nn),pTimeStep)] = 0

    #%% generation parameters
    pGenToTechnology  = dfGeneration['Technology'     ]                                                                            # generator association to technology
    pGenToCompany     = dfGeneration['Company'        ]                                                                            # generator association to company
    pRatedMinPower    = dfGeneration['MinimumPower'   ] * 1e-3                                                                     # rated minimum power                 [GW]
    pRatedMaxPower    = dfGeneration['MaximumPower'   ] * 1e-3                                                                     # rated maximum power                 [GW]
    pLinearVarCost    = dfGeneration['LinearTerm'     ] * 1e-3 * dfGeneration['FuelCost'] + dfGeneration['OMVariableCost'] * 1e-3  # linear   term variable cost         [MEUR/GWh]
    pConstantVarCost  = dfGeneration['ConstantTerm'   ] * 1e-6 * dfGeneration['FuelCost']                                          # constant term variable cost         [MEUR/h]
    pStartUpCost      = dfGeneration['StartUpCost'    ]                                                                            # startup  cost                       [MEUR]
    pShutDownCost     = dfGeneration['ShutDownCost'   ]                                                                            # shutdown cost                       [MEUR]
    pRampUp           = dfGeneration['RampUp'         ] * 1e-3                                                                     # ramp up   rate                      [GW/h]
    pRampDw           = dfGeneration['RampDown'       ] * 1e-3                                                                     # ramp down rate                      [GW/h]
    pCO2EmissionRate  = dfGeneration['CO2EmissionRate'] * 1e-3                                                                     # emission  rate                      [t CO2/MWh]
    pUpTime           = dfGeneration['UpTime'         ]                                                                            # minimum up   time                   [h]
    pDwTime           = dfGeneration['DownTime'       ]                                                                            # minimum down time                   [h]
    pMaxCharge        = dfGeneration['MaximumCharge'  ] * 1e-3                                                                     # maximum ESS charge                  [GW]
    pInitialInventory = dfGeneration['InitialStorage' ]                                                                            # initial ESS storage                 [GWh]
    pRatedMinStorage  = dfGeneration['MinimumStorage' ]                                                                            # minimum ESS storage                 [GWh]
    pRatedMaxStorage  = dfGeneration['MaximumStorage' ]                                                                            # maximum ESS storage                 [GWh]
    pEfficiency       = dfGeneration['Efficiency'     ]                                                                            #         ESS efficiency              [p.u.]
    pStorageType      = dfGeneration['StorageType'    ]                                                                            #         ESS type

    ReadingDataTime = time.time() - StartTime
    StartTime       = time.time()
    print('Reading    input data                 ... ', round(ReadingDataTime), 's')

    #%% defining subsets: active load levels (n), thermal units (t), ESS units (es), all the lines (la), candidate lines (lc), and lines with losses (ll)
    mSDUC.n  = Set(doc='load levels'     , initialize=[nn for nn in mSDUC.nn if pDuration     [nn] >  0  ])
    mSDUC.n2 = Set(doc='load levels'     , initialize=[nn for nn in mSDUC.nn if pDuration     [nn] >  0  ])
    mSDUC.g  = Set(doc='generating units', initialize=[gg for gg in mSDUC.gg if pRatedMaxPower[gg] >  0.0])
    mSDUC.t  = Set(doc='thermal    units', initialize=[g  for g  in mSDUC.g  if pLinearVarCost [g] >  0.0])
    mSDUC.r  = Set(doc='RES        units', initialize=[g  for g  in mSDUC.g  if pLinearVarCost [g] == 0.0 and pRatedMaxStorage[g] == 0.0])
    mSDUC.es = Set(doc='ESS        units', initialize=[g  for g  in mSDUC.g  if                               pRatedMaxStorage[g] >  0.0])

    # non-RES units
    mSDUC.nr = mSDUC.g - mSDUC.r

    if pTimeStep > 1:
        # drop levels with duration 0
        pDuration           = pDuration.loc          [mSDUC.sc*mSDUC.n]
        pDemand             = pDemand.loc            [mSDUC.sc*mSDUC.n]
        pOperReserveUp      = pOperReserveUp.loc     [mSDUC.sc*mSDUC.n]
        pOperReserveDw      = pOperReserveDw.loc     [mSDUC.sc*mSDUC.n]
        pVariableMinPower   = pVariableMinPower.loc  [mSDUC.sc*mSDUC.n]
        pVariableMaxPower   = pVariableMaxPower.loc  [mSDUC.sc*mSDUC.n]
        pVariableMinStorage = pVariableMinStorage.loc[mSDUC.sc*mSDUC.n]
        pVariableMaxStorage = pVariableMaxStorage.loc[mSDUC.sc*mSDUC.n]
        pEnergyInflows      = pEnergyInflows.loc     [mSDUC.sc*mSDUC.n]

    # variable minimum and maximum power
    pVariableMinPower = pVariableMinPower.replace(0.0, float('nan'))
    pVariableMaxPower = pVariableMaxPower.replace(0.0, float('nan'))
    pMinPower         = pd.DataFrame([pRatedMinPower]*len(pVariableMaxPower.index), index=pd.Index(pVariableMaxPower.index), columns=pRatedMinPower.index)
    pMaxPower         = pd.DataFrame([pRatedMaxPower]*len(pVariableMaxPower.index), index=pd.Index(pVariableMaxPower.index), columns=pRatedMaxPower.index)
    pMinPower         = pMinPower.reindex        (sorted(pMinPower.columns        ), axis=1)
    pMaxPower         = pMaxPower.reindex        (sorted(pMaxPower.columns        ), axis=1)
    pVariableMinPower = pVariableMinPower.reindex(sorted(pVariableMinPower.columns), axis=1)
    pVariableMaxPower = pVariableMaxPower.reindex(sorted(pVariableMaxPower.columns), axis=1)
    pMinPower         = pVariableMinPower.where(pVariableMinPower > pMinPower, other=pMinPower)
    pMaxPower         = pVariableMaxPower.where(pVariableMaxPower < pMaxPower, other=pMaxPower)
    pMaxPower2ndBlock = pMaxPower - pMinPower

    # variable minimum and maximum storage capacity
    pVariableMinStorage = pVariableMinStorage.replace(0.0, float('nan'))
    pVariableMaxStorage = pVariableMaxStorage.replace(0.0, float('nan'))
    pMinStorage         = pd.DataFrame([pRatedMinStorage]*len(pVariableMinStorage.index), index=pd.Index(pVariableMinStorage.index), columns=pRatedMinStorage.index)
    pMaxStorage         = pd.DataFrame([pRatedMaxStorage]*len(pVariableMaxStorage.index), index=pd.Index(pVariableMaxStorage.index), columns=pRatedMaxStorage.index)
    pMinStorage         = pMinStorage.reindex        (sorted(pMinStorage.columns        ), axis=1)
    pMaxStorage         = pMaxStorage.reindex        (sorted(pMaxStorage.columns        ), axis=1)
    pVariableMinStorage = pVariableMinStorage.reindex(sorted(pVariableMinStorage.columns), axis=1)
    pVariableMaxStorage = pVariableMaxStorage.reindex(sorted(pVariableMaxStorage.columns), axis=1)
    pMinStorage         = pVariableMinStorage.where(pVariableMinStorage > pMinStorage, other=pMinStorage)
    pMaxStorage         = pVariableMaxStorage.where(pVariableMaxStorage < pMaxStorage, other=pMaxStorage)

    # values < 1e-5 times the maximum system demand are converted to 0
    pEpsilon = pDemand.max()*1e-5
    # these parameters are in GW
    pDemand          [pDemand           < pEpsilon] = 0.0
    pOperReserveUp   [pOperReserveUp    < pEpsilon] = 0.0
    pOperReserveDw   [pOperReserveDw    < pEpsilon] = 0.0
    pMinPower        [pMinPower         < pEpsilon] = 0.0
    pMaxPower        [pMaxPower         < pEpsilon] = 0.0
    pMaxPower2ndBlock[pMaxPower2ndBlock < pEpsilon] = 0.0
    pMaxCharge       [pMaxCharge        < pEpsilon] = 0.0
    pEnergyInflows   [pEnergyInflows    < pEpsilon/pTimeStep] = 0.0
    # these parameters are in GWh
    pMinStorage      [pMinStorage       < pEpsilon] = 0.0
    pMaxStorage      [pMaxStorage       < pEpsilon] = 0.0

    # this option avoids a warning in the following assignments
    pd.options.mode.chained_assignment = None

    # minimum up- and downtime converted to an integer number of time steps
    pUpTime = round(pUpTime/pTimeStep).astype('int')
    pDwTime = round(pDwTime/pTimeStep).astype('int')

    # thermal and variable units ordered by increasing variable cost
    mSDUC.go = pLinearVarCost.sort_values().index

    # determine the initial committed units and their output
    pInitialOutput = pd.Series([0.0]*len(mSDUC.g), dfGeneration.index)
    pInitialUC     = pd.Series([0.0]*len(mSDUC.g), dfGeneration.index)
    pSystemOutput  = 0.0
    for go in mSDUC.go:
        n1 = next(iter(mSDUC.sc*mSDUC.n))
        if pSystemOutput < pDemand[n1]:
            if go in mSDUC.r:
                pInitialOutput[go] = pMaxPower[go][n1]
            else:
                pInitialOutput[go] = pMinPower[go][n1]
            pInitialUC    [go] = 1
            pSystemOutput     += pInitialOutput[go]

    #%% variables
    mSDUC.vTotalVCost     = Var(                                    within=NonNegativeReals, doc='total system variable cost [MEUR]')
    mSDUC.vTotalECost     = Var(                                    within=NonNegativeReals, doc='total system emission cost [MEUR]')
    mSDUC.vTotalOutput    = Var(mSDUC.sc*mSDUC.n*mSDUC.g , within=NonNegativeReals, doc='total output of the unit     [GW]')
    mSDUC.vOutput2ndBlock = Var(mSDUC.sc*mSDUC.n*mSDUC.nr, within=NonNegativeReals, doc='second block of the unit     [GW]')
    mSDUC.vReserveUp      = Var(mSDUC.sc*mSDUC.n*mSDUC.nr, within=NonNegativeReals, doc='operating reserve up         [GW]')
    mSDUC.vReserveDown    = Var(mSDUC.sc*mSDUC.n*mSDUC.nr, within=NonNegativeReals, doc='operating reserve down       [GW]')
    mSDUC.vESSInventory   = Var(mSDUC.sc*mSDUC.n*mSDUC.es, within=NonNegativeReals, doc='ESS inventory               [GWh]')
    mSDUC.vESSSpillage    = Var(mSDUC.sc*mSDUC.n*mSDUC.es, within=NonNegativeReals, doc='ESS spillage                [GWh]')
    mSDUC.vESSCharge      = Var(mSDUC.sc*mSDUC.n*mSDUC.es, within=NonNegativeReals, doc='ESS    charge power          [GW]')
    mSDUC.vENS            = Var(mSDUC.sc*mSDUC.n,          within=NonNegativeReals, doc='energy not served in node    [GW]')

    mSDUC.vCommitment     = Var(         mSDUC.n*mSDUC.nr, within=Binary,           doc='commitment of the unit      {0,1}')
    mSDUC.vStartUp        = Var(         mSDUC.n*mSDUC.nr, within=Binary,           doc='StartUp    of the unit      {0,1}')
    mSDUC.vShutDown       = Var(         mSDUC.n*mSDUC.nr, within=Binary,           doc='ShutDown   of the unit      {0,1}')

    [mSDUC.vTotalOutput   [sc,n,g ].setub(pMaxPower.loc        [(sc,n),g ]) for sc,n,g  in mSDUC.sc*mSDUC.n*mSDUC.g ]
    [mSDUC.vOutput2ndBlock[sc,n,nr].setub(pMaxPower2ndBlock.loc[(sc,n),nr]) for sc,n,nr in mSDUC.sc*mSDUC.n*mSDUC.nr]
    [mSDUC.vReserveUp     [sc,n,nr].setub(pMaxPower2ndBlock.loc[(sc,n),nr]) for sc,n,nr in mSDUC.sc*mSDUC.n*mSDUC.nr]
    [mSDUC.vReserveDown   [sc,n,nr].setub(pMaxPower2ndBlock.loc[(sc,n),nr]) for sc,n,nr in mSDUC.sc*mSDUC.n*mSDUC.nr]
    [mSDUC.vESSInventory  [sc,n,es].setlb(pMinStorage.loc      [(sc,n),es]) for sc,n,es in mSDUC.sc*mSDUC.n*mSDUC.es]
    [mSDUC.vESSInventory  [sc,n,es].setub(pMaxStorage.loc      [(sc,n),es]) for sc,n,es in mSDUC.sc*mSDUC.n*mSDUC.es]
    [mSDUC.vESSCharge     [sc,n,es].setub(pMaxCharge.loc       [       es]) for sc,n,es in mSDUC.sc*mSDUC.n*mSDUC.es]
    [mSDUC.vENS           [sc,n   ].setub(pDemand.loc          [ sc,n    ]) for sc,n    in mSDUC.sc*mSDUC.n         ]

    # fixing the ESS inventory at the last load level at the end of the time scope
    for sc,es in mSDUC.sc*mSDUC.es:
        mSDUC.vESSInventory[sc,mSDUC.n.last(),es].fix(pInitialInventory[es])

    #%% definition of the time-steps leap to observe the stored energy at ESS
    pCycleTimeStep = pUpTime*0
    for es in mSDUC.es:
        if  pStorageType[es] == 'Daily'  :
            pCycleTimeStep[es] =        1
        if  pStorageType[es] == 'Weekly'  :
            pCycleTimeStep[es] = int(  24/pTimeStep)
        if  pStorageType[es] == 'Monthly' :
            pCycleTimeStep[es] = int( 168/pTimeStep)

    # fixing the ESS inventory at the end of the following pCycleTimeStep (weekly, yearly), i.e., for daily ESS is fixed at the end of the week, for weekly/monthly ESS is fixed at the end of the year
    for sc,n,es in mSDUC.sc*mSDUC.n*mSDUC.es:
         if pStorageType[es] == 'Daily'   and mSDUC.n.ord(n) % ( 168/pTimeStep) == 0:
             mSDUC.vESSInventory[sc,n,es].fix(pInitialInventory[es])
         if pStorageType[es] == 'Weekly'  and mSDUC.n.ord(n) % ( 672/pTimeStep) == 0:
             mSDUC.vESSInventory[sc,n,es].fix(pInitialInventory[es])
         if pStorageType[es] == 'Monthly' and mSDUC.n.ord(n) % (8736/pTimeStep) == 0:
             mSDUC.vESSInventory[sc,n,es].fix(pInitialInventory[es])

    SettingUpDataTime = time.time() - StartTime
    StartTime         = time.time()
    print('Setting up input data                 ... ', round(SettingUpDataTime), 's')

    def eTotalVCost(mSDUC):
        return mSDUC.vTotalVCost == (sum(pScenProb[sc] * pDuration[n] * pENSCost             * mSDUC.vENS        [sc,n   ] for sc,n    in mSDUC.sc*mSDUC.n         ) +
                                     sum(pScenProb[sc] * pDuration[n] * pLinearVarCost  [nr] * mSDUC.vTotalOutput[sc,n,nr] for sc,n,nr in mSDUC.sc*mSDUC.n*mSDUC.nr) +
                                     sum(                pDuration[n] * pConstantVarCost[nr] * mSDUC.vCommitment [   n,nr]                                           +
                                                                        pStartUpCost    [nr] * mSDUC.vStartUp    [   n,nr]                                           +
                                                                        pShutDownCost   [nr] * mSDUC.vShutDown   [   n,nr] for    n,nr in          mSDUC.n*mSDUC.nr) )
    mSDUC.eTotalVCost = Constraint(rule=eTotalVCost, doc='total system variable cost [MEUR]')

    def eTotalECost(mSDUC):
        return mSDUC.vTotalECost == sum(pScenProb[sc] * pCO2Cost * pCO2EmissionRate[nr] * mSDUC.vTotalOutput[sc,n,nr] for sc,n,nr in mSDUC.sc*mSDUC.n*mSDUC.nr)
    mSDUC.eTotalECost = Constraint(rule=eTotalECost, doc='total system emission cost [MEUR]')

    def eTotalTCost(mSDUC):
        return mSDUC.vTotalVCost + mSDUC.vTotalECost
    mSDUC.eTotalTCost = Objective(rule=eTotalTCost, sense=minimize, doc='total system cost [MEUR]')

    GeneratingOFTime = time.time() - StartTime
    StartTime        = time.time()
    print('Generating objective function         ... ', round(GeneratingOFTime), 's')

    #%% constraints
    def eOperReserveUp(mSDUC,sc,n):
        if pScenProb[sc] and pOperReserveUp[sc,n]:
            return sum(mSDUC.vReserveUp  [sc,n,nr] for nr in mSDUC.nr) >= pOperReserveUp[sc,n]
        else:
            return Constraint.Skip
    mSDUC.eOperReserveUp = Constraint(mSDUC.sc*mSDUC.n, rule=eOperReserveUp, doc='up   operating reserve [GW]')

    def eOperReserveDw(mSDUC,sc,n):
        if pScenProb[sc] and pOperReserveDw[sc,n]:
            return sum(mSDUC.vReserveDown[sc,n,nr] for nr in mSDUC.nr) >= pOperReserveDw[sc,n]
        else:
            return Constraint.Skip
    mSDUC.eOperReserveDw = Constraint(mSDUC.sc*mSDUC.n, rule=eOperReserveDw, doc='down operating reserve [GW]')

    def eBalance(mSDUC,sc,n):
        if pScenProb[sc]:
            return sum(mSDUC.vTotalOutput[sc,n,g] for g in mSDUC.g) - sum(mSDUC.vESSCharge[sc,n,es] for es in mSDUC.es) + mSDUC.vENS[sc,n] == pDemand[sc,n]
        else:
            return Constraint.Skip
    mSDUC.eBalance = Constraint(mSDUC.sc*mSDUC.n, rule=eBalance, doc='load generation balance [GW]')

    def eESSInventory(mSDUC,sc,n,es):
        if   pScenProb[sc] and mSDUC.n.ord(n) == pCycleTimeStep[es]:
            return pInitialInventory[es]                                         + sum(pDuration[n2]*(pEnergyInflows[es][sc,n2] - mSDUC.vTotalOutput[sc,n2,es] + pEfficiency[es]*mSDUC.vESSCharge[sc,n2,es]) for n2 in list(mSDUC.n2)[mSDUC.n.ord(n)-pCycleTimeStep[es]:mSDUC.n.ord(n)]) == mSDUC.vESSInventory[sc,n,es] + mSDUC.vESSSpillage[sc,n,es]
        elif pScenProb[sc] and mSDUC.n.ord(n) >  pCycleTimeStep[es] and mSDUC.n.ord(n) % pCycleTimeStep[es] == 0:
            return mSDUC.vESSInventory[sc,mSDUC.n.prev(n,pCycleTimeStep[es]),es] + sum(pDuration[n2]*(pEnergyInflows[es][sc,n2] - mSDUC.vTotalOutput[sc,n2,es] + pEfficiency[es]*mSDUC.vESSCharge[sc,n2,es]) for n2 in list(mSDUC.n2)[mSDUC.n.ord(n)-pCycleTimeStep[es]:mSDUC.n.ord(n)]) == mSDUC.vESSInventory[sc,n,es] + mSDUC.vESSSpillage[sc,n,es]
        else:
            return Constraint.Skip
    mSDUC.eESSInventory = Constraint(mSDUC.sc*mSDUC.n*mSDUC.es, rule=eESSInventory, doc='ESS inventory balance [GWh]')

    GeneratingRBITime = time.time() - StartTime
    StartTime         = time.time()
    print('Generating reserves/balance/inventory ... ', round(GeneratingRBITime), 's')

    #%%
    def eMaxOutput2ndBlock(mSDUC,sc,n,nr):
        if   pScenProb[sc] and pOperReserveUp[sc,n] and pMaxPower2ndBlock[nr][sc,n]:
            return (mSDUC.vOutput2ndBlock[sc,n,nr] + mSDUC.vReserveUp  [sc,n,nr]) / pMaxPower2ndBlock[nr][sc,n] <= mSDUC.vCommitment[n,nr]
        else:
            return Constraint.Skip
    mSDUC.eMaxOutput2ndBlock = Constraint(mSDUC.sc*mSDUC.n*mSDUC.nr, rule=eMaxOutput2ndBlock, doc='max output of the second block of a committed unit [p.u.]')

    def eMinOutput2ndBlock(mSDUC,sc,n,nr):
        if   pScenProb[sc] and pOperReserveDw[sc,n] and pMaxPower2ndBlock[nr][sc,n]:
            return (mSDUC.vOutput2ndBlock[sc,n,nr] + mSDUC.vReserveDown[sc,n,nr]) / pMaxPower2ndBlock[nr][sc,n] >= 0.0
        else:
            return Constraint.Skip
    mSDUC.eMinOutput2ndBlock = Constraint(mSDUC.sc*mSDUC.n*mSDUC.nr, rule=eMinOutput2ndBlock, doc='min output of the second block of a committed unit [p.u.]')

    def eTotalOutput(mSDUC,sc,n,nr):
        if   pScenProb[sc] and pMinPower[nr][sc,n] == 0.0:
            return mSDUC.vTotalOutput[sc,n,nr]                       ==                           mSDUC.vOutput2ndBlock[sc,n,nr]
        elif pScenProb[sc]:
            return mSDUC.vTotalOutput[sc,n,nr] / pMinPower[nr][sc,n] == mSDUC.vCommitment[n,nr] + mSDUC.vOutput2ndBlock[sc,n,nr] / pMinPower[nr][sc,n]
        else:
            return Constraint.Skip
    mSDUC.eTotalOutput = Constraint(mSDUC.sc*mSDUC.n*mSDUC.nr, rule=eTotalOutput, doc='total output of a unit [GW]')

    def eUCStrShut(mSDUC,n,nr):
        if   n == mSDUC.n.first():
            return mSDUC.vCommitment[n,nr] - pInitialUC[nr]                        == mSDUC.vStartUp[n,nr] - mSDUC.vShutDown[n,nr]
        else:
            return mSDUC.vCommitment[n,nr] - mSDUC.vCommitment[mSDUC.n.prev(n),nr] == mSDUC.vStartUp[n,nr] - mSDUC.vShutDown[n,nr]
    mSDUC.eUCStrShut = Constraint(mSDUC.n*mSDUC.nr, rule=eUCStrShut, doc='relation among commitment startup and shutdown')

    GeneratingGenConsTime = time.time() - StartTime
    StartTime             = time.time()
    print('Generating generation constraints     ... ', round(GeneratingGenConsTime), 's')

    #%%
    def eRampUp(mSDUC,sc,n,t):
        if   pScenProb[sc] and pRampUp[t] and pRampUp[t] < pMaxPower2ndBlock[t][sc,n] and n == mSDUC.n.first():
            return (mSDUC.vOutput2ndBlock[sc,n,t] - max(pInitialOutput[t]-pMinPower[t][sc,n],0.0) + mSDUC.vReserveUp  [sc,n,t]) / pDuration[n] / pRampUp[t] <= mSDUC.vCommitment[n,t] - mSDUC.vStartUp[n,t]
        elif pScenProb[sc] and pRampUp[t] and pRampUp[t] < pMaxPower2ndBlock[t][sc,n]:
            return (mSDUC.vOutput2ndBlock[sc,n,t] - mSDUC.vOutput2ndBlock[sc,mSDUC.n.prev(n),t]   + mSDUC.vReserveUp  [sc,n,t]) / pDuration[n] / pRampUp[t] <= mSDUC.vCommitment[n,t] - mSDUC.vStartUp[n,t]
        else:
            return Constraint.Skip
    mSDUC.eRampUp = Constraint(mSDUC.sc*mSDUC.n*mSDUC.t, rule=eRampUp, doc='maximum ramp up   [p.u.]')

    def eRampDw(mSDUC,sc,n,t):
        if   pScenProb[sc] and pRampDw[t] and pRampDw[t] < pMaxPower2ndBlock[t][sc,n] and n == mSDUC.n.first():
            return (mSDUC.vOutput2ndBlock[sc,n,t] - max(pInitialOutput[t]-pMinPower[t][sc,n],0.0) - mSDUC.vReserveDown[sc,n,t]) / pDuration[n] / pRampDw[t] >= - pInitialUC[t]                        + mSDUC.vShutDown[n,t]
        elif pScenProb[sc] and pRampDw[t] and pRampDw[t] < pMaxPower2ndBlock[t][sc,n]:
            return (mSDUC.vOutput2ndBlock[sc,n,t] - mSDUC.vOutput2ndBlock[sc,mSDUC.n.prev(n),t]   - mSDUC.vReserveDown[sc,n,t]) / pDuration[n] / pRampDw[t] >= - mSDUC.vCommitment[mSDUC.n.prev(n),t] + mSDUC.vShutDown[n,t]
        else:
            return Constraint.Skip
    mSDUC.eRampDw = Constraint(mSDUC.sc*mSDUC.n*mSDUC.t, rule=eRampDw, doc='maximum ramp down [p.u.]')

    GeneratingRampsTime = time.time() - StartTime
    StartTime           = time.time()
    print('Generating ramps   up/down            ... ', round(GeneratingRampsTime), 's')

    #%%
    def eMinUpTime(mSDUC,n,t):
        if pScenProb[sc] and pUpTime[t] > 1 and mSDUC.n.ord(n) >= pUpTime[t]:
            return sum(mSDUC.vStartUp [n2,t] for n2 in list(mSDUC.n2)[mSDUC.n.ord(n)-pUpTime[t]:mSDUC.n.ord(n)]) <=     mSDUC.vCommitment[n,t]
        else:
            return Constraint.Skip
    mSDUC.eMinUpTime   = Constraint(mSDUC.n*mSDUC.t, rule=eMinUpTime  , doc='minimum up   time [h]')

    def eMinDownTime(mSDUC,n,t):
        if pScenProb[sc] and pDwTime[t] > 1 and mSDUC.n.ord(n) >= pDwTime[t]:
            return sum(mSDUC.vShutDown[n2,t] for n2 in list(mSDUC.n2)[mSDUC.n.ord(n)-pDwTime[t]:mSDUC.n.ord(n)]) <= 1 - mSDUC.vCommitment[n,t]
        else:
            return Constraint.Skip
    mSDUC.eMinDownTime = Constraint(mSDUC.n*mSDUC.t, rule=eMinDownTime, doc='minimum down time [h]')

    GeneratingMinUDTime = time.time() - StartTime
    StartTime           = time.time()
    print('Generating minimum up/down time       ... ', round(GeneratingMinUDTime), 's')

    #%% solving the problem
    mSDUC.write(_path+'/openSDUC_'+CaseName+'.lp', io_options={'symbolic_solver_labels': True}) # create lp-format file
    Solver = SolverFactory(SolverName)                                                   # select solver
    if SolverName == 'gurobi':
        Solver.options['LogFile'       ] = _path+'/openSDUC_'+CaseName+'.log'
        #Solver.options['IISFile'      ] = 'openSDUC_'+CaseName+'.ilp'                   # should be uncommented to show results of IIS
        #Solver.options['Method'       ] = 2                                             # barrier method
        Solver.options['MIPGap'        ] = 0.02
        Solver.options['Threads'       ] = int((psutil.cpu_count(logical=True) + psutil.cpu_count(logical=False))/2)
        #Solver.options['TimeLimit'     ] = 7200
        #Solver.options['IterationLimit'] = 7200000
    SolverResults = Solver.solve(mSDUC, tee=True)                                        # tee=True displays the output of the solver
    SolverResults.write()                                                                # summary of the solver results

    #%% fix values of binary variables to get dual variables and solve it again
    for n,nr in mSDUC.n*mSDUC.nr:
        mSDUC.vCommitment[n,nr].fix(round(mSDUC.vCommitment[n,nr]()))
        mSDUC.vStartUp   [n,nr].fix(round(mSDUC.vStartUp   [n,nr]()))
        mSDUC.vShutDown  [n,nr].fix(round(mSDUC.vShutDown  [n,nr]()))

    mSDUC.dual    = Suffix(direction=Suffix.IMPORT)
    SolverResults = Solver.solve(mSDUC, tee=True)                                        # tee=True displays the output of the solver
    SolverResults.write()                                                                # summary of the solver results

    SolvingTime = time.time() - StartTime
    StartTime   = time.time()
    print('Solving                               ... ', round(SolvingTime), 's')

    print('Objective function value                  ', mSDUC.eTotalTCost())

    #%% inverse index generator to technology and to company
    pTechnologyToGen = pGenToTechnology.reset_index().set_index('Technology').set_axis(['Generator'], axis=1, copy=False)[['Generator']]
    pTechnologyToGen = pTechnologyToGen.loc[pTechnologyToGen['Generator'].isin(mSDUC.g)]
    pTechnology2Gen  = pTechnologyToGen.reset_index().set_index(['Technology', 'Generator'])

    mSDUC.t2g = Set(initialize=pTechnology2Gen.index, doc='technology to generator')

    #%% outputting the generation operation

    OutputResults = pd.Series(data=[mSDUC.vCommitment[n,nr]() for n,nr in mSDUC.n*mSDUC.nr], index=pd.Index(mSDUC.n*mSDUC.nr))
    OutputResults.to_frame(name='p.u.').reset_index().pivot_table(index=['level_0'], columns='level_1', values='p.u.').rename_axis(['LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_GenerationCommitment_'+CaseName+'.csv', sep=',')
    OutputResults = pd.Series(data=[mSDUC.vStartUp   [n,nr]() for n,nr in mSDUC.n*mSDUC.nr], index=pd.Index(mSDUC.n*mSDUC.nr))
    OutputResults.to_frame(name='p.u.').reset_index().pivot_table(index=['level_0'], columns='level_1', values='p.u.').rename_axis(['LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_GenerationStartUp_'   +CaseName+'.csv', sep=',')
    OutputResults = pd.Series(data=[mSDUC.vShutDown  [n,nr]() for n,nr in mSDUC.n*mSDUC.nr], index=pd.Index(mSDUC.n*mSDUC.nr))
    OutputResults.to_frame(name='p.u.').reset_index().pivot_table(index=['level_0'], columns='level_1', values='p.u.').rename_axis(['LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_GenerationShutDown_'  +CaseName+'.csv', sep=',')

    sSCN   = [(sc,n   ) for sc,n    in mSDUC.sc*mSDUC.n          if pScenProb[sc]]
    sSCNG  = [(sc,n,g ) for sc,n,g  in mSDUC.sc*mSDUC.n*mSDUC.g  if pScenProb[sc]]
    sSCNNR = [(sc,n,nr) for sc,n,nr in mSDUC.sc*mSDUC.n*mSDUC.nr if pScenProb[sc]]
    sSCNES = [(sc,n,es) for sc,n,es in mSDUC.sc*mSDUC.n*mSDUC.es if pScenProb[sc]]
    sSCNR  = [(sc,n,r ) for sc,n,r  in mSDUC.sc*mSDUC.n*mSDUC.r  if pScenProb[sc]]
    sSCNT  = [(sc,n,t ) for sc,n,t  in mSDUC.sc*mSDUC.n*mSDUC.t  if pScenProb[sc]]
    sSCNGT = [(sc,n,gt) for sc,n,gt in mSDUC.sc*mSDUC.n*mSDUC.gt if pScenProb[sc]]
    if sum(pOperReserveUp[sc,n] for sc,n in sSCN):
        OutputResults = pd.Series(data=[mSDUC.vReserveUp  [sc,n,nr]()*1e3 for sc,n,nr in sSCNNR], index=pd.Index(sSCNNR))
        OutputResults.to_frame(name='MW').reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='MW').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_GenerationReserveUp_'+CaseName+'.csv', sep=',')

    if sum(pOperReserveDw[sc,n] for sc,n in sSCN):
        OutputResults = pd.Series(data=[mSDUC.vReserveDown[sc,n,nr]()*1e3 for sc,n,nr in sSCNNR], index=pd.Index(sSCNNR))
        OutputResults.to_frame(name='MW').reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='MW').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_GenerationReserveDown_'+CaseName+'.csv', sep=',')

    OutputResults = pd.Series(data=[mSDUC.vTotalOutput[sc,n,g]()*1e3      for sc,n,g in sSCNG], index=pd.Index(sSCNG))
    OutputResults.to_frame(name='MW').reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='MW').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_Generation_'+CaseName+'.csv', sep=',')

    OutputResults = pd.Series(data=[mSDUC.vENS[sc,n]()*1e3                for sc,n in sSCN], index=pd.Index(sSCN))
    OutputResults.to_frame(name='MW').reset_index().pivot_table(index=['level_0','level_1'], values='MW').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_PNS_'+CaseName+'.csv', sep=',')

    OutputResults = pd.Series(data=[mSDUC.vENS[sc,n]()*pDuration[n]       for sc,n in sSCN], index=pd.Index(sSCN))
    OutputResults.to_frame(name='GWh').reset_index().pivot_table(index=['level_0','level_1'], values='GWh').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_ENS_'+CaseName+'.csv', sep=',')

    OutputResults = pd.Series(data=[(mSDUC.vTotalOutput[sc,n,g].ub-mSDUC.vTotalOutput[sc,n,g]())*1e3 for sc,n,g in sSCNR], index=pd.Index(sSCNR))
    OutputResults.to_frame(name='MW').reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='MW').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_GenerationCurtailment_'+CaseName+'.csv', sep=',')

    #%% plot SRMC for all the scenarios
    RESCurtailment = OutputResults.loc[:,:,:]

    fig, fg = plt.subplots()
    for r in mSDUC.r:
        fg.plot(range(len(sSCN)), RESCurtailment[:,:,r], label=r)
    fg.set(xlabel='Hours', ylabel='MW')
    fg.set_ybound(lower=0)
    plt.title('RES Curtailment')
    fg.tick_params(axis='x', rotation=90)
    fg.legend()
    plt.tight_layout()
    #plt.show()
    plt.savefig(_path+'/oUC_Plot_RESCurtailment_'+CaseName+'.png', bbox_inches=None)

    OutputResults = pd.Series(data=[mSDUC.vTotalOutput[sc,n,g]()*pDuration[n]                                  for sc,n,g in sSCNG], index=pd.Index(sSCNG))
    OutputResults.to_frame(name='GWh' ).reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='GWh' ).rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_GenerationEnergy_'+CaseName+'.csv', sep=',')

    OutputResults = pd.Series(data=[mSDUC.vTotalOutput[sc,n,nr]()*pCO2EmissionRate[nr]*1e3                     for sc,n,nr in sSCNT], index=pd.Index(sSCNT))
    OutputResults.to_frame(name='tCO2').reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='tCO2').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_GenerationEmission_'+CaseName+'.csv', sep=',')

    #%% outputting the ESS operation
    if len(mSDUC.es):
        OutputResults = pd.Series(data=[mSDUC.vESSCharge[sc,n,es]()*1e3                                        for sc,n,es in sSCNES], index=pd.Index(sSCNES))
        OutputResults.to_frame(name='MW' ).reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='MW' ).rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_Consumption_'+CaseName+'.csv', sep=',')

        OutputResults = pd.Series(data=[sum(OutputResults[sc,n,es] for es in mSDUC.es if (gt,es) in mSDUC.t2g) for sc,n,gt in sSCNGT if sum(1 for es in mSDUC.es if (gt,es) in mSDUC.t2g)], index=pd.Index([(sc,n,gt) for sc,n,gt in sSCNGT if sum(1 for es in mSDUC.es if (gt,es) in mSDUC.t2g)]))
        OutputResults.to_frame(name='MW' ).reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='MW' ).rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_TechnologyConsumption_'+CaseName+'.csv', sep=',')

        TechnologyCharge = OutputResults.loc[:,:,:]

        OutputResults = pd.Series(data=[mSDUC.vESSCharge[sc,n,es]()*pDuration[n]                               for sc,n,es in sSCNES], index=pd.Index(sSCNES))
        OutputResults.to_frame(name='GWh').reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='GWh').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_ConsumptionEnergy_'+CaseName+'.csv', sep=',')

        OutputResults = pd.Series(data=[sum(OutputResults[sc,n,es] for es in mSDUC.es if (gt,es) in mSDUC.t2g) for sc,n,gt in sSCNGT if sum(1 for es in mSDUC.es if (gt,es) in mSDUC.t2g)], index=pd.Index([(sc,n,gt) for sc,n,gt in sSCNGT if sum(1 for es in mSDUC.es if (gt,es) in mSDUC.t2g)]))
        OutputResults.to_frame(name='GWh').reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='GWh').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_TechnologyConsumptionEnergy_'+CaseName+'.csv', sep=',')

        OutputResults = pd.Series(data=[mSDUC.vESSInventory[sc,n,es]()                                         for sc,n,es in sSCNES], index=pd.Index(sSCNES))
        OutputResults.to_frame(name='GWh').reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='GWh', dropna=False).rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_ESSInventory_'+CaseName+'.csv', sep=',')

        OutputResults = pd.Series(data=[mSDUC.vESSSpillage [sc,n,es]()                                         for sc,n,es in sSCNES], index=pd.Index(sSCNES))
        OutputResults.to_frame(name='GWh').reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='GWh', dropna=False).rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_ESSSpillage_'+CaseName+'.csv', sep=',')

    OutputResults = pd.Series(data=[mSDUC.vTotalOutput[sc,n,g]()*1e3                                           for sc,n,g  in sSCNG ], index=pd.Index(sSCNG))
    OutputResults = pd.Series(data=[sum(OutputResults[sc,n,g] for g  in mSDUC.g  if (gt,g ) in mSDUC.t2g)      for sc,n,gt in sSCNGT if sum(1 for g in mSDUC.g if (gt,g) in mSDUC.t2g)], index=pd.Index([(sc,n,gt) for sc,n,gt in sSCNGT if sum(1 for g in mSDUC.g if (gt,g) in mSDUC.t2g)]))
    OutputResults.to_frame(name='MW').reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='MW').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_TechnologyGeneration_'+CaseName+'.csv', sep=',')

    TechnologyOutput = OutputResults.loc[:,:,:]

    # for sc in mSDUC.sc:
    #     fig, fg = plt.subplots()
    #     fg.stackplot(range(len(mSDUC.n)),  TechnologyOutput.loc[sc,:,:].values.reshape(len(mSDUC.n),len(mSDUC.gt)).transpose().tolist(), baseline='zero', labels=list(mSDUC.gt))
    #     fg.plot     (range(len(mSDUC.n)), -TechnologyCharge.loc[sc,:,'ESS'], label='ESSCharge', linewidth=0.5, color='b')
    #     fg.plot     (range(len(mSDUC.n)),  pDemand[sc]*1e3,                  label='Demand'   , linewidth=0.5, color='k')
    #     fg.set(xlabel='Hours', ylabel='MW')
    #     plt.title(sc)
    #     fg.tick_params(axis='x', rotation=90)
    #     fg.legend()
    #     plt.tight_layout()
    #     #plt.show()
    #     plt.savefig(_path+'/oUC_Plot_Technology_'+sc+'_'+CaseName+'.png', bbox_inches=None)

    OutputResults = pd.Series(data=[mSDUC.vTotalOutput[sc,n,g]()*pDuration[n]                          for sc,n,g  in sSCNG], index=pd.Index(sSCNG))
    OutputResults = pd.Series(data=[sum(OutputResults[sc,n,g] for g in mSDUC.g if (gt,g) in mSDUC.t2g) for sc,n,gt in sSCNGT if sum(1 for g in mSDUC.g if (gt,g) in mSDUC.t2g)], index=pd.Index([(sc,n,gt) for sc,n,gt in sSCNGT if sum(1 for g in mSDUC.g if (gt,g) in mSDUC.t2g)]))
    OutputResults.to_frame(name='GWh').reset_index().pivot_table(index=['level_0','level_1'], columns='level_2', values='GWh').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_TechnologyGenerationEnergy_'+CaseName+'.csv', sep=',')

    #%% outputting the SRMC
    if SolverName == 'gurobi' or SolverName == 'appsi_highs':
        OutputResults = pd.Series(data=[mSDUC.dual[mSDUC.eBalance[sc,n]]*1e3/pScenProb[sc]/pDuration[n] for sc,n in sSCN], index=pd.Index(sSCN))
        OutputResults.to_frame(name='SRMC').reset_index().pivot_table(index=['level_0','level_1'], values='SRMC').rename_axis(['Scenario','LoadLevel'], axis=0).rename_axis([None], axis=1).to_csv(_path+'/oUC_Result_SRMC_'+CaseName+'.csv', sep=',')

        #%% plot SRMC for all the scenarios
        SRMC = OutputResults.loc[:,:]

        fig, fg = plt.subplots()
        for sc in mSDUC.sc:
            if pScenProb[sc]:
                fg.plot(range(len(mSDUC.n)), SRMC[sc], label=sc)
                fg.set(xlabel='Hours', ylabel='EUR/MWh')
                fg.set_ybound(lower=0, upper=100)
                plt.title('SRMC')
                fg.tick_params(axis='x', rotation=90)
                fg.legend()
                plt.tight_layout()
                #plt.show()
                plt.savefig(_path+'/oUC_Plot_SRMC_'+CaseName+'.png', bbox_inches=None)

    WritingResultsTime = time.time() - StartTime
    StartTime          = time.time()
    print('Writing output results                ... ', round(WritingResultsTime), 's')
    print('Total time                            ... ', round(ReadingDataTime + GeneratingOFTime + GeneratingRBITime + GeneratingGenConsTime + GeneratingRampsTime + GeneratingMinUDTime + SolvingTime + WritingResultsTime), 's')
    print('\033[34m#### Academic research license - for non-commercial use only ####\033[0m')

if __name__ == '__main__':
    main()
